local addonName, addonTable = ...
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local icon = ldb and LibStub("LibDBIcon-1.0",true)
local WWLDB = nil
local C_Timer = C_Timer
local C_Garrison = C_Garrison
local TAKE_ACTION_NOTHING, TAKE_ACTION_WORK, TAKE_ACTION_SLEEP = 0, 1, 2
local currentRealmChar -- Make Option be character specific

local BARRACKS_IDS = { [26] = true, [27] = true, [28] = true }
local GARDEN_IDS   = { [29] = true, [136] = true, [137] = true }


local function dbg(...)
  --[===[@debug@
  if not WorkWork_SV.profile[currentRealmChar].shutTheFuckUp then
    DEFAULT_CHAT_FRAME:AddMessage("|cff2476ffDebugWorkWork:|r " .. string.join(", ", tostringall(...)))
  end
  --@end-debug@]===]
end

local function say(text)
  if WorkWork_SV.profile[currentRealmChar].shutTheFuckUp then
    return
  end
  DEFAULT_CHAT_FRAME:AddMessage("|cff2476ffWorkWork:|r " .. text)
end


local WorkWork = CreateFrame("Frame")

--[===[@debug@
WWDBG = WorkWork
--@end-debug@]===]



WorkWork.Events = {}
WorkWork.StopAllTimers = nil
WorkWork.formatStrings = {}
WorkWork.formatStrings.icon = "|T%s:16:16:0:0:64:64:5:59:5:59|t"
WorkWork:RegisterEvent("ADDON_LOADED")
WorkWork:RegisterEvent("GARRISON_ARCHITECT_OPENED")
WorkWork:RegisterEvent("GARRISON_ARCHITECT_CLOSED")
WorkWork:RegisterEvent("GARRISON_UPDATE")
WorkWork:SetScript("OnEvent",function( self, event, ...)
  self.Events[event](self, ...)
end)


local function GetTipAnchor(frame)
  local x, y = frame:GetCenter()
  if not x or not y then return "TOPLEFT", "BOTTOMLEFT" end
  local hhalf = (x > UIParent:GetWidth() * 2 / 3) and "RIGHT" or (x < UIParent:GetWidth() / 3) and "LEFT" or ""
  local vhalf = (y > UIParent:GetHeight() / 2) and "TOP" or "BOTTOM"
  return vhalf .. hhalf, frame, (vhalf == "TOP" and "BOTTOM" or "TOP") .. hhalf
end

WorkWork.Events["ADDON_LOADED"] = function( self, loadedAddon )
  if loadedAddon ~= addonName then return end
  self:UnregisterEvent("ADDON_LOADED")

  currentRealmChar = string.format("%s.%s", GetRealmName(), UnitName("player"))
  self:UpgradeDB()
  -- LDB

  WWLDB = ldb:NewDataObject("WorkWork", {
    type = "data source",
    label = "WorkWork",
    icon = "Interface\\Icons\\inv_mace_10",
    OnClick = function(self, button)
      WorkWork:ButtonClicked( button )
      WorkWork:ShowHoverTooltip( self )
    end,
    OnEnter = function(frame)
      WorkWork:ShowHoverTooltip(frame)
    end,
    OnLeave = function(self)
      GameTooltip:Hide()
    end,
  })

  -- Button and Feedback checkbox on architect table frame
  self:CreateButtons()

  -- Update icon and text to the current state
  icon:Register("WorkWork", WWLDB, WorkWork_SV.profile[currentRealmChar].minimapIcon)

  dbg("loaded")
end

WorkWork.Events["GARRISON_UPDATE"] = function( self )
  self:UnregisterEvent("GARRISON_UPDATE")
  C_Timer.After( 1, function() WorkWork:TimerCallback() end )
end

WorkWork.Events["GARRISON_ARCHITECT_OPENED"] = function( self )
  self.StopAllTimers = nil
  WorkWork:StartWorking()
end

WorkWork.TimerCallback = function( self )
  if WorkWork_SV.profile[currentRealmChar].TakeAction == TAKE_ACTION_WORK then
    -- Smack those peons and get them to work
    self:SmackPeons()
  elseif WorkWork_SV.profile[currentRealmChar].TakeAction == TAKE_ACTION_SLEEP then
    -- Sing a lullaby, making peons sleepy and leaving work
    self:SingLullaby()
  end
end

WorkWork.Events["GARRISON_ARCHITECT_CLOSED"] = function( self )
  self.StopAllTimers = true
end

function WorkWork:EnDisableStatus(checkAgainst)
  if not checkAgainst then
    return "Enable"
  else
    return "Disable"
  end
end

function WorkWork:ShowHoverTooltip(anchorFrame)
  local sv = WorkWork_SV.profile[currentRealmChar]
  GameTooltip:SetOwner(anchorFrame, "ANCHOR_CURSOR")
  GameTooltip:ClearLines()
  GameTooltip:AddLine( "      ", 1, 1, 1)
  GameTooltip:AddLine( WWLDB.text, 1, 1, 1)
  GameTooltip:AddLine( "Left-Click: Cycle modes", 1, 1, 1)
  GameTooltip:AddLine( string.format("Shift Left-Click: %s Chatframe ouput",     self:EnDisableStatus(not sv.shutTheFuckUp)), 1, 1, 1)
  GameTooltip:AddLine( string.format("Right-Click: %s Barracks ignore",          self:EnDisableStatus(sv.skipBarracks     )), 1, 1, 1)
  GameTooltip:AddLine( string.format("Shift Right-Click: %s Herb Garden ignore", self:EnDisableStatus(sv.skipHerb         )), 1, 1, 1)
  GameTooltip:AddLine( string.format("Middle-Click: %s Automatic Mode",          self:EnDisableStatus(sv.automatic        )), 1, 1, 1)
  GameTooltip:ClearAllPoints()
  GameTooltip:SetClampedToScreen(true)
  GameTooltip:SetPoint(GetTipAnchor(anchorFrame))
  GameTooltip:Show()
end

function WorkWork:StartWorking()
  if not GarrisonBuildingFrame:IsVisible() then
    return
  end

  -- Automatic mode:
  -- Opening table while not in  sleepmode and previous order is done
  -- => Switch to the opposite mode and start it
  if WorkWork_SV.profile[currentRealmChar].automatic and
     WorkWork_SV.profile[currentRealmChar].TakeAction ~= TAKE_ACTION_NOTHING and
     self.WorkDoneFeedbackCheckbox:GetChecked() then
    local switchTo = ( WorkWork_SV.profile[currentRealmChar].TakeAction == TAKE_ACTION_WORK ) and TAKE_ACTION_SLEEP or TAKE_ACTION_WORK
    WorkWork:CycleStates( switchTo )
  end
  C_Timer.After( 0.2, function() WorkWork:TimerCallback() end )
end

WorkWork.buildings = {}
WorkWork.buildingsWithWorkers = {}

WorkWork.followers = {}
WorkWork.followersList = {}
-- C_Garrison.AssignFollowerToBuilding(GarrisonBuildingFrame.selectedBuilding.plotID, self.info.followerID);
-- C_Garrison.GetBuildings() to get a table of BuildingID <=> PlotID
WorkWork.SortFollowers = GarrisonFollowerList_SortFollowers
function WorkWork:SmackPeons()
  dbg("SmackPeons")
  self.buildings = C_Garrison.GetBuildings()
  local are_we_done = true

  -- Got buildings to work with
  self.buildingsWithWorkers = {}
  if #self.buildings > 0 then
    for i = 1, #self.buildings do
      -- Only search the followers if the building is not occupied yet
      -- Fuckton of returns, we only care if it has follower slot and
      -- if it is currently being build
      local _, _, _, texture, _, _, _, _, _, _, _, _, _, _, _, _, hasFollowerSlot, _, _, _, isBeingBuilt, _, _, _, canCompleteBuild = C_Garrison.GetOwnedBuildingInfo(self.buildings[i].plotID)
      if not ( C_Garrison.GetFollowerInfoForBuilding(self.buildings[i].plotID) ) then
        -- Skip Barracks/Herb Garden option
        -- Skip buildings that are being build at the moment
        if not ( BARRACKS_IDS[self.buildings[i].buildingID] and WorkWork_SV.profile[currentRealmChar].skipBarracks ) and
           not ( GARDEN_IDS[self.buildings[i].buildingID] and WorkWork_SV.profile[currentRealmChar].skipHerb ) and
           not isBeingBuilt and not canCompleteBuild then
          -- Get a list of ALL follower who can work, includes inactives and ones currently on missions!
          self.followers = C_Garrison.GetPossibleFollowersForBuilding(self.buildings[i].plotID);
          if (#self.followers > 0) then
            -- Sorts the followers like wow would from best to worst
            self.SortFollowers(self)

            for j = 1, #self.followers do
              -- Use the first one thats available
              -- C_Garrison.GetFollowerStatus returns nil for available
              if not C_Garrison.GetFollowerStatus(self.followers[j].followerID) then
                dbg(self.followers[j].name .. " (" ..self.followers[j].followerID..") => " ..self.buildings[i].texPrefix .." (" ..self.buildings[i].plotID..")")
                self:GoToWork(self.buildings[i].plotID, self.followers[j].followerID)
                are_we_done = false
              end -- C_Garrison.GetFollowerStatus
            end -- for followers
          end -- skiprax?
        end -- followers > 0
      else
        table.insert(self.buildingsWithWorkers, string.format(self.formatStrings.icon, texture))
      end -- someone already working?
    end -- for buildings


    self:UpdateLDB( are_we_done )
    if are_we_done == true then
      -- All possible work slots filled
      -- self:UpdateLDB( true )
      say("Smacked all the peons, everyone is working")
    else
      -- We tried to assign at least one worker, as long as the architect table was not closed in the meantime go for a nother run in 0.1s
      if not self.StopAllTimers or not GarrisonBuildingFrame:IsVisible() then
        self:StartWorking()
      end
    end
  end
end

function WorkWork:GoToWork( plot, follower )
  C_Garrison.AssignFollowerToBuilding( plot, follower )
end


-- To clear all buildings of their followers we just need to iterate over all the buildings.
function WorkWork:SingLullaby()
  self.buildings = C_Garrison.GetBuildings()
  local are_we_done = true
  if #self.buildings > 0 then
    for i = 1, #self.buildings do
      if C_Garrison.GetFollowerInfoForBuilding( self.buildings[i].plotID ) and
         not( BARRACKS_IDS[self.buildings[i].buildingID] and WorkWork_SV.profile[currentRealmChar].skipBarracks ) and
         not( GARDEN_IDS[self.buildings[i].buildingID] and WorkWork_SV.profile[currentRealmChar].skipHerb ) then
        self:GoToSleep( self.buildings[i].plotID )
        are_we_done = false
      end
    end -- for buildings
  end -- buildings > 0


  if are_we_done then
    self:UpdateLDB( true )
    say("Lullaby sung, all peons recalled from their workplaces")
  else
    if not self.StopAllTimers then
      self:StartWorking()
    end
  end
end

function WorkWork:GoToSleep( plot )
  C_Garrison.RemoveFollowerFromBuilding( plot )
end

SLASH_WORKWORK1, SLASH_WORKWORK2 = '/workwork', '/ww';

local STATES = {
  [TAKE_ACTION_NOTHING] = "No actions taken",
  [TAKE_ACTION_WORK]    = "Smacking peons",
  [TAKE_ACTION_SLEEP]   = "Singing Lullaby",
}
function WorkWork:CycleStates( force )
  WorkWork_SV.profile[currentRealmChar].TakeAction = force or ( ( WorkWork_SV.profile[currentRealmChar].TakeAction + 1 ) % 3 )

  self:UpdateLDB()
  say(STATES[WorkWork_SV.profile[currentRealmChar].TakeAction])
  return WorkWork_SV.profile[currentRealmChar].TakeAction
end

function WorkWork:UpdateLDB( done )
  local icn = "Interface\\AddOns\\WorkWork\\media\\workwork_"..WorkWork_SV.profile[currentRealmChar].TakeAction
  WWLDB.icon = icn

  local doneIcon = done and string.format(self.formatStrings.icon, "Interface\\Buttons\\UI-CheckBox-Check") or ""

  self.GarrisonBuildingFrameButton:SetNormalTexture( icn )
  self.WorkDoneFeedbackCheckbox:SetChecked( done )
  if WorkWork_SV.profile[currentRealmChar].TakeAction == TAKE_ACTION_NOTHING then
    self:UnregisterEvent("GARRISON_ARCHITECT_OPENED")
    WWLDB.text = "Stopped"
  elseif WorkWork_SV.profile[currentRealmChar].TakeAction == TAKE_ACTION_WORK then
    self:RegisterEvent("GARRISON_ARCHITECT_OPENED")
    WWLDB.text = table.concat(self.buildingsWithWorkers, " ") .. doneIcon
    self.ft:SetText( WWLDB.text )
  elseif WorkWork_SV.profile[currentRealmChar].TakeAction == TAKE_ACTION_SLEEP then
    self:RegisterEvent("GARRISON_ARCHITECT_OPENED")
    WWLDB.text = "zZzZZ" .. doneIcon
    self.ft:SetText( WWLDB.text )
  else
    dbg("This should not happen!")
    WorkWork_SV.profile[currentRealmChar].TakeAction = TAKE_ACTION_NOTHING
  end

  if done then
    GarrisonBuildingFrame_UpdatePlots()
  end
end

function WorkWork:ButtonClicked( button )
  if button == "RightButton" then
    if IsShiftKeyDown() then
      WorkWork:SkipGarden()
    else
      WorkWork:SkipBarracks()
    end
  elseif button == "MiddleButton" then
    WorkWork:ToggleAutoMode()
  else
    if IsShiftKeyDown() then
      WorkWork:Stfu()
    else
      WorkWork:CycleStates()
      WorkWork:StartWorking()
    end
  end
end

function WorkWork:CreateButtons()
  local toggle = CreateFrame("CheckButton", "WorkWorkDoneFeedbackCheckbox", GarrisonBuildingFrame, "InterfaceOptionsCheckButtonTemplate")
  toggle:SetSize(24, 24) toggle:SetHitRectInsets(0,0,0,0)
  toggle:SetPoint("RIGHT", GarrisonBuildingFrame.CloseButton, "Left", 5, 0)
  toggle:Disable()
  self.WorkDoneFeedbackCheckbox = toggle

  local btn =  CreateFrame("Button", "WorkWorkGarrisonBuildingFrameButton", GarrisonBuildingFrame)
  btn:SetNormalTexture( "Interface\\AddOns\\WorkWork\\media\\workwork_"..WorkWork_SV.profile[currentRealmChar].TakeAction )
  btn:SetSize(24, 24)
  btn:SetHitRectInsets(0,0,0,0)
  btn:SetPoint("RIGHT", toggle, "LEFT", -5, 0)
  self.GarrisonBuildingFrameButton = btn
  btn:RegisterForClicks("AnyDown")
  btn:SetScript("OnClick", function(self, button, down)
    WorkWork:ButtonClicked( button )
    WorkWork:ShowHoverTooltip( self )
  end)
  btn:SetScript("OnEnter", function( frame )
    WorkWork:ShowHoverTooltip(frame)
  end)
  btn:SetScript("OnLeave", function( self )
    GameTooltip:Hide()
  end)
  local ft = GarrisonBuildingFrame:CreateFontString("WorkWorkBuildingFrameText", "OVERLAY ", "GameFontHighlight")
  ft:SetSize(240, 20)
  ft:SetPoint("RIGHT", btn, "LEFT")
  ft:SetJustifyH("RIGHT")
  --ft:SetText()
  self.ft=ft
end

function WorkWork:Stfu()
  say("Disabled chatframe output")
  WorkWork_SV.profile[currentRealmChar].shutTheFuckUp = not WorkWork_SV.profile[currentRealmChar].shutTheFuckUp
  say("Enabled chatframe output")
end

function WorkWork:ToggleAutoMode()
    WorkWork_SV.profile[currentRealmChar].automatic = not WorkWork_SV.profile[currentRealmChar].automatic
    say("Automatic mode ".. (WorkWork_SV.profile[currentRealmChar].automatic and "enabled" or "disabled") )
end

function WorkWork:SkipBarracks()
  WorkWork_SV.profile[currentRealmChar].skipBarracks = not WorkWork_SV.profile[currentRealmChar].skipBarracks
  if WorkWork_SV.profile[currentRealmChar].skipBarracks  then
    say("Barracks ignored")
  else
    say("Removed Barracks ignore")
  end
end

function WorkWork:SkipGarden()
  WorkWork_SV.profile[currentRealmChar].skipHerb = not WorkWork_SV.profile[currentRealmChar].skipHerb
  if WorkWork_SV.profile[currentRealmChar].skipHerb  then
    say("Herb Garden ignored")
  else
    say("Removed Herb Garden ignore")
  end
end

function WorkWork:UpgradeDB()

  WorkWork_SV = WorkWork_SV or {}
  WorkWork_SV.profile = WorkWork_SV.profile or {}
  WorkWork_SV.global  = WorkWork_SV.global or {}
  if not WorkWork_SV.global.DB_VERSION then
    -- Old SV was not per character
    -- Upgrade
    WorkWork_SV.global.DB_VERSION = 1
  end
  -- Switch to character specific settings
  if not WorkWork_SV.profile[currentRealmChar] then
    WorkWork_SV.profile[currentRealmChar] = {}

    WorkWork_SV.profile[currentRealmChar].minimapIcon = {}

    WorkWork_SV.profile[currentRealmChar].TakeAction = WorkWork_SV.TakeAction or TAKE_ACTION_WORK
    if WorkWork_SV.skipBarracks == nil then
      WorkWork_SV.profile[currentRealmChar].skipBarracks = true
    end
    if WorkWork_SV.skipHerb == nil then
      WorkWork_SV.profile[currentRealmChar].skipHerb = true
    end

    WorkWork_SV.profile[currentRealmChar].hide          = WorkWork_SV.hide or false
    WorkWork_SV.profile[currentRealmChar].shutTheFuckUp = WorkWork_SV.shutTheFuckUp or false
    WorkWork_SV.profile[currentRealmChar].automatic     = WorkWork_SV.automatic or true
  end
end

local function handler(msg, editbox)
  if msg == "stfu" then
    WorkWork:Stfu()
  elseif msg == "show" then
    WorkWork_SV.profile[currentRealmChar].minimapIcon.hide = false
    icon:Show("WorkWork")
  elseif msg == "auto" then
    WorkWork:ToggleAutoMode()
  elseif msg == "hide" then
    WorkWork_SV.profile[currentRealmChar].minimapIcon.hide = true
    icon:Hide("WorkWork")
  elseif msg == "skiprax" then
    WorkWork:SkipBarracks()
  elseif msg == "skipherb" then
    WorkWork:SkipGarden()
  else
    WorkWork:CycleStates()
  end
end
SlashCmdList["WORKWORK"] = handler;